import controlP5.*;

//size of simulation
int W = 800;
int H = 700;
float heights[][] = new float[W][H];
float heightsNew[][] = new float[W][H];

//scaling factor for drawing sand levels
float maxHeight=20;
float minHeight=0;

//simulation variable
float windStr = .5;//amount of sand that is transported
float gravStr = .1;//strength of diffusion/gravity
float windTrans = 2.0;//distance sand travels
float epsilon = .3;//

//hold previous and next values of for periodic boundary conditions
int pI[] = new int[W];
int nI[] = new int[W];
int pJ[] = new int[H];
int nJ[] = new int[H];

void setup() {
  size(W,H);
  //load GUI to control simulation variables
  setupGUI();
  //setup periodic boundary conditions
  for(int i=0;i<W;++i) {
    pI[i] = (i-1+W)%W;
    nI[i] = (i+1)%W;
  }
  for(int i=0;i<H;++i) {
    pJ[i] = (i-1+H)%H;
    nJ[i] = (i+1)%H;
  }

  initHeights(); 
}

//initialize heights, some base value with noise
void initHeights() {
  for(int i=0;i<W;++i) {
    for(int j=0;j<H;++j) {
      heights[i][j] = 5+random(.0001);
    }
  }
}

void draw() {
  update();
  draw2D();
}

//draw heights as a greyscale image
void draw2D() {
  camera();
  loadPixels();
  for(int i=0;i<W;++i) {
    for(int j=0;j<H;++j) {
      color c = color((heights[i][j]-minHeight)/(maxHeight-minHeight)*255);
      pixels[j*width+i] = c;
    }
  }
  updatePixels();
}

//update the sand heights by the Nishi Mori algorithm
//this is the main part
void update() {
  //initialize variables so we aren't making new ones inside the loop
  float moveGrav, moveWind, grad;
  float windDist;
  int windDistI; //integer wind distance for linear interpolation
  float temp[][]; //temporary variable for swapping my new heights and old heights
  
  //new height starts at the old height value
  for(int i=0;i<W;++i) {
    for(int j=0;j<H;++j) {
      heightsNew[i][j] = heights[i][j];
    }
  }
  
  //movement due to gravity (downhill) this acts like diffusion
  for(int i=0;i<W;++i) {
    for(int j=0;j<H;++j) {
      //five point laplacian, average the neighbors and subtract self
      moveGrav = heights[pI[i]][j]+heights[nI[i]][j]+heights[i][pJ[j]]+heights[i][nJ[j]]-4.0*heights[i][j];
      moveGrav *= gravStr;
      heightsNew[i][j] += moveGrav;
    }
  }
  //swap new heights and old heights
  //this simply switches pointers so it is very fast
  temp = heights;
  heights = heightsNew;
  heightsNew = temp;

  for(int i=0;i<W;++i) {
    for(int j=0;j<H;++j) {
      heightsNew[i][j] = heights[i][j];
    }
  }

  //"saltation step" moving sand due to wind
  for(int i=0;i<W;++i) {
    for(int j=0;j<H;++j) {
      //gradient in the direction of the wind (+x direction)
      grad = constrain(heights[nI[i]][j]-heights[i][j],-3,3);
      
      //approximate tanh function (x*(27+x^2)/(27+9*x^2))
      grad = grad*(27.0+grad*grad)/(27.0+9.0*grad*grad);

      //amount of sand that gets moved
      moveWind = windStr*(-grad+1+epsilon);
      
      //make sure that amount is not greater than the amount that is there
      if(moveWind>heights[i][j]) moveWind = heights[i][j];
      //distance that sand moves
      windDist = windTrans*(grad+1);
      //subtract sand from current 
      heightsNew[i][j] -= moveWind;

      //linearly interpolate transported sand to neighboring cells
      windDistI = int(windDist);
      if(windDistI == windDist) {
        heightsNew[(i+windDistI)%W][j] += moveWind;
      } 
      else {
        heightsNew[(i+windDistI)%W][j] += (windDist-windDistI)*moveWind;
        heightsNew[(i+windDistI+1)%W][j] += (1-(windDist-windDistI))*moveWind;
      }
    }
  }
  
  //swap new and old heights
  temp = heights;
  heights = heightsNew;
  heightsNew = temp;
}

